import doctest
from zope.testing import renormalizing

import logging
import os
import re
import shutil
import stat
import tempfile
import unittest
import zc.buildout.testing
from zc.buildout.testing import buildoutTearDown
from functools import wraps
from subprocess import check_call, check_output, CalledProcessError
from slapos.recipe.gitclone import GIT_CLONE_ERROR_MESSAGE, \
    GIT_CLONE_CACHE_ERROR_MESSAGE

optionflags = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)

GIT_REPOSITORY = 'https://lab.nexedi.com/nexedi/slapos.recipe.build.git'
BAD_GIT_REPOSITORY = 'http://git.erp5.org/repos/nowhere'
REVISION = '2566127'

def setUp(test):
  # XXX side effect. Disable libnetworkcache because buildout testing
  # infrastructure doesn't support offline install of external egg "forced"
  # into itself.
  zc.buildout.buildout.LIBNETWORKCACHE_ENABLED = False

  zc.buildout.testing.buildoutSetUp(test)
  zc.buildout.testing.install_develop('slapos.recipe.build', test)

def with_buildout(wrapped):
  def wrapper(self):
    self.globs = {}
    setUp(self)
    try:
      wrapped(self, **self.globs)
    finally:
      buildoutTearDown(self)
  return wraps(wrapped)(wrapper)

class GitCloneNonInformativeTests(unittest.TestCase):

  def setUp(self):
    self.dir = os.path.realpath(tempfile.mkdtemp())
    self.parts_directory_path = os.path.join(self.dir, 'test_parts')

  def setUpSubmoduleRepository(self):
    """
    This function sets up repositories and config for parent repo and a
    submodule repo and create small commits as well as links the 2 repos.
    """
    # Create parent and submodule directory
    self.project_dir = os.path.join(self.dir, 'main_repo')
    self.submodule_dir = os.path.join(self.dir ,'submodule_repo')
    os.mkdir(self.project_dir)
    os.mkdir(self.submodule_dir)

    # Add files in submodule repo and initialize git in it
    check_call(['git', 'init'], cwd=self.submodule_dir)
    self.setUpGitConfig(self.submodule_dir)
    self.touch(self.submodule_dir, 'file1.py')
    self.gitAdd(self.submodule_dir)
    self.gitCommit(self.submodule_dir, msg='Add file1 in submodule repo')

    # Add files and folder in parent repo and initialize git in it
    check_call(['git', 'init'], cwd=self.project_dir)
    self.setUpGitConfig(self.project_dir)
    submodule_dir_main_repo = os.path.join(self.project_dir, 'dir1')
    os.mkdir(submodule_dir_main_repo)
    self.touch(self.project_dir, 'file1.py')
    self.gitAdd(self.project_dir)
    self.gitCommit(self.project_dir, msg='Add file and folder in main repo')

    # Add submodule to main repo and commit
    check_call(['git', 'submodule', 'add', self.submodule_dir],
                cwd=submodule_dir_main_repo)
    self.gitCommit(self.project_dir, msg='Add submodule repo')

  def tearDown(self):
    shutil.rmtree(self.dir)
    for var in list(os.environ):
      if var.startswith('SRB_'):
        del os.environ[var]

  def write_file(self, filename, contents, mode=stat.S_IREAD | stat.S_IWUSR):
    path = os.path.join(self.dir, filename)
    fh = open(path, 'w')
    fh.write(contents)
    fh.close()
    os.chmod(path, mode)
    return path

  def setUpGitConfig(self, proj_dir):
    """
    Setup user and email for given git repository
    """
    check_call(['git', 'config', 'user.email', 'test@example.com'], cwd=proj_dir)
    check_call(['git', 'config', 'user.name', 'Test'], cwd=proj_dir)

  def gitAdd(self, proj_dir):
    """runs a 'git add .' in the provided git repo
    :param proj_dir:    path to a git repo
    """
    try:
      check_call(['git', 'add', '.'], cwd=proj_dir)
    except CalledProcessError as e:
      logging.error("'git add .' failed with '%s'" % e.returncode)

  def gitCommit(self, proj_dir, msg):
    """runs a 'git commit -m msg' in the provided git repo
    :param proj_dir:  path to a git repo
    """
    try:
      check_call(['git', 'commit', '-m', msg], cwd=proj_dir)
    except CalledProcessError as e:
      logging.error("'git commit' failed with '%s'" % e.returncode)

  def getRepositoryHeadCommit(self, proj_dir):
    """
    Returns the sha of HEAD of a git repo
    """
    try :
      output = check_output(['git', 'rev-parse', 'HEAD'],
                            cwd=proj_dir).decode("utf-8")
      return output.rstrip()
    except CalledProcessError:
      logging.error("failed to call 'git rev-parse'")
      return None

  def touch(self, *parts):
    os.close(os.open(os.path.join(*parts), os.O_CREAT, 0o666))

  def echo(self, path, line):
    myFile = open(path, 'w')
    myFile.write(line)
    myFile.close()

  def makeGitCloneRecipe(self, options):
    from slapos.recipe.gitclone import Recipe
    bo = {
        'buildout': {
            'parts-directory': self.parts_directory_path,
            'directory': self.dir,
         }
    }
    default_options = {
      'repository': GIT_REPOSITORY
      }
    default_options.update(**options)
    return Recipe(bo, 'test', default_options)

  def test_using_download_cache_if_git_fails(self):
    recipe = self.makeGitCloneRecipe({"use-cache": "true",
                                      "repository": BAD_GIT_REPOSITORY})
    os.chdir(self.dir)
    with self.assertRaises(zc.buildout.UserError) as cm:
      recipe.install()
    self.assertEqual(str(cm.exception), GIT_CLONE_CACHE_ERROR_MESSAGE)

  def test_not_using_download_cache_if_forbidden(self):
    recipe = self.makeGitCloneRecipe({"repository": BAD_GIT_REPOSITORY})
    os.chdir(self.dir)
    with self.assertRaises(zc.buildout.UserError) as cm:
      recipe.install()
    self.assertEqual(str(cm.exception), GIT_CLONE_ERROR_MESSAGE)

  def test_cleanup_of_pyc_files(self):
    recipe = self.makeGitCloneRecipe({})
    recipe.install()
    git_repository_path = os.path.join(self.parts_directory_path, "test")
    self.assertTrue(os.path.exists(git_repository_path))
    bad_file_path = os.path.join(git_repository_path, "foo.pyc")
    bade_file = open(bad_file_path, 'w')
    bade_file.close()
    self.assertTrue(os.path.exists(bad_file_path))
    # install again and make sure pyc file is removed
    recipe = self.makeGitCloneRecipe({})
    recipe.update()
    self.assertTrue(os.path.exists(git_repository_path))
    self.assertFalse(os.path.exists(bad_file_path), "pyc file not removed")

  @with_buildout
  def test_clone_and_update_submodule(self, buildout, write, sample_buildout,
                                      **kw):
    """
    STEP I:
    Repositories status:  Main repo(M1) and Submodule repo (S1)
    Main repo (M1) ---references---> Submodule(S1)
    Install should be M1+S1

    STEP II:
    Repositories status:  Main repo(M2) and Submodule repo (S2)
    Main repo (M2) ---references---> Submodule(S1)
    Install should be M2+S1

    STEP III:
    Repositories status:  Main repo(M3) and Submodule repo (S2)
    Main repo (M3) ---references---> Submodule(S2)
    Install should be M3+S2
    """
    self.setUpSubmoduleRepository()

    # STEP I: Clone repositories in status M1 and S1 (M1---->S1)
    write(sample_buildout, 'buildout.cfg',
"""
[buildout]
parts = git-clone

[git-clone]
recipe = slapos.recipe.build:gitclone
repository = %s
""" % self.project_dir)
    check_call([buildout])

    main_repo_path = os.path.join(sample_buildout, 'parts', 'git-clone')
    self.assertTrue(os.path.exists(main_repo_path))
    submodule_repo_path = os.path.join(main_repo_path, 'dir1',
                                       'submodule_repo')
    # Check if the submodule is not empty
    self.assertTrue(os.listdir(submodule_repo_path))

    # Get the head commit of the submodule repo
    head_commit_submodule_after_clone = self.getRepositoryHeadCommit(
                                              submodule_repo_path)

    # Add a marker file in parent as well as submodule repo
    parent_local_change_path = os.path.join(main_repo_path,
                                            'local_change_main')
    submodule_local_change_path = os.path.join(submodule_repo_path,
                                               'local_change_submodule')
    self.echo(parent_local_change_path, 'foo')
    self.echo(submodule_local_change_path, 'bar')

    # Check if the file are created at the expected position and check contents
    self.assertTrue(os.path.exists(parent_local_change_path))
    self.assertTrue(os.path.exists(submodule_local_change_path))
    self.assertEqual(check_output(['cat', parent_local_change_path]), 'foo')
    self.assertEqual(check_output(['cat', submodule_local_change_path]), 'bar')

    # Update buildout
    check_call([buildout])

    # The local changes should be still there
    self.assertTrue(os.path.exists(parent_local_change_path))
    self.assertTrue(os.path.exists(submodule_local_change_path))
    self.assertEqual(check_output(['cat', parent_local_change_path]), 'foo')
    self.assertEqual(check_output(['cat', submodule_local_change_path]), 'bar')

    # Update submodule repository and main repo separately
    self.touch(self.project_dir, 'file2.py')
    self.gitAdd(self.project_dir)
    self.gitCommit(self.project_dir, msg='Add file2 in main repo')
    self.touch(self.submodule_dir, 'file2.py')
    self.gitAdd(self.submodule_dir)
    self.gitCommit(self.submodule_dir, msg='Add file2 in submodule repo')

    # STEP II: Clone repositories in status M2 and S2 (M2---->S1)
    # Update the recipe with new revision which still point to the submodule
    # old revision
    write(sample_buildout, 'buildout.cfg',
"""
[buildout]
parts = git-clone

[git-clone]
recipe = slapos.recipe.build:gitclone
repository = %s
revision = %s
""" % (self.project_dir, str(self.getRepositoryHeadCommit(self.project_dir))))
    check_call([buildout])

    self.assertTrue(os.path.exists(main_repo_path))
    # Check if the submodule is not empty
    self.assertTrue(os.listdir(submodule_repo_path))

    # Since calling buildout should've reinstalled, we expect the local changes
    # to be gone
    self.assertFalse(os.path.exists(parent_local_change_path))
    self.assertFalse(os.path.exists(submodule_local_change_path))

    # Get the head commit of the submodule repo
    head_commit_submodule_after_first_revision = self.getRepositoryHeadCommit(
                                                      submodule_repo_path)

    # Change the HEAD of the cloned repository and then run buildout
    check_call(['git', 'reset', '--hard', 'HEAD'], cwd=main_repo_path)
    # Update buildout. This should fetch and reset to the revision given in
    # the buildout
    check_call([buildout])
    head_commit_parent_repo_after_update = self.getRepositoryHeadCommit(
                                                  main_repo_path)
    self.assertEqual(head_commit_parent_repo_after_update,
                     self.getRepositoryHeadCommit(self.project_dir))

    # Update main repo with changes of submodule
    submodule_dir_main_repo = os.path.join(self.project_dir, 'dir1',
                                           'submodule_repo')
    check_call(['git', 'checkout', 'master'], cwd=submodule_dir_main_repo)
    check_call(['git', 'pull', '--ff'], cwd=submodule_dir_main_repo)
    self.gitAdd(self.project_dir)
    self.gitCommit(self.project_dir, msg='Update submodule version')

    # STEP III: Clone repositories in status M3 and S2 (M3---->S2)
    # Update the recipe with new revision which points to submodule new revision
    write(sample_buildout, 'buildout.cfg',
"""
[buildout]
parts = git-clone

[git-clone]
recipe = slapos.recipe.build:gitclone
repository = %s
revision = %s
""" % (self.project_dir, str(self.getRepositoryHeadCommit(self.project_dir))))
    check_call([buildout])

    self.assertTrue(os.path.exists(main_repo_path))
    # Check if the submodule is not empty
    self.assertTrue(os.listdir(submodule_repo_path))
    # Get the head commit of the submodule repo
    head_commit_submodule_after_second_revision = self.getRepositoryHeadCommit(
                                                      submodule_repo_path)

    # Check the HEAD of the submodule
    submodule_head_commit = self.getRepositoryHeadCommit(self.submodule_dir)
    self.assertEqual(head_commit_submodule_after_clone,
                     head_commit_submodule_after_first_revision)

    self.assertNotEqual(head_commit_submodule_after_first_revision,
                        head_commit_submodule_after_second_revision)

    self.assertEqual(head_commit_submodule_after_second_revision,
                     submodule_head_commit)

  def test_ignore_ssl_certificate(self, ignore_ssl_certificate=True):
    import slapos.recipe.gitclone

    # Monkey patch check_call
    original_check_call = slapos.recipe.gitclone.check_call
    check_call_parameter_list = []
    def patch_check_call(*args, **kw):
      check_call_parameter_list.extend((args, kw))
      original_check_call(args[0])
    slapos.recipe.gitclone.check_call = patch_check_call

    bo = {
        'buildout': {
            'parts-directory': self.parts_directory_path,
            'directory': self.dir,
         }
    }
    options = {
      'repository': GIT_REPOSITORY,
      "ignore-ssl-certificate": str(ignore_ssl_certificate).lower(),
    }
    recipe = slapos.recipe.gitclone.Recipe(bo, 'test', options)

    recipe.install()

    # Check git clone parameters
    _ = self.assertIn if ignore_ssl_certificate else self.assertNotIn
    _("http.sslVerify=false", check_call_parameter_list[0][0])

    # Restore original check_call method
    slapos.recipe.gitclone.check_call = original_check_call

  def test_ignore_ssl_certificate_false(self):
    self.test_ignore_ssl_certificate(ignore_ssl_certificate=False)

  def test_clone_submodules_by_default(self, ignore_cloning_submodules=False):
    self.setUpSubmoduleRepository()
    recipe = self.makeGitCloneRecipe(
      {'repository': self.project_dir,
        'ignore-cloning-submodules': str(ignore_cloning_submodules).lower()}
    )
    recipe.install()
    main_repo_path = os.path.join(self.parts_directory_path, "test")
    self.assertTrue(os.path.exists(main_repo_path))
    submodule_repo_path = os.path.join(main_repo_path, 'dir1',
                                       'submodule_repo')
    # Check if the folder exists
    self.assertTrue(os.path.exists(main_repo_path))
    # Check is there is anything in submodule repository path
    self.assertNotEqual(bool(ignore_cloning_submodules),
                        bool(os.listdir(submodule_repo_path)))

  def test_ignore_cloning_submodules(self):
    self.test_clone_submodules_by_default(ignore_cloning_submodules=True)

  def test_fetch_submodules_with_main_repo(self):
    """
    Test to check the fetch of submodules while fetching main repo. Updating
    should udpate the main repo as well as submodule repo if the reference of
    submodule in main repo has been udpated.
    """
    self.setUpSubmoduleRepository()
    recipe = self.makeGitCloneRecipe(
      {'repository': self.project_dir}
    )
    recipe.install()
    main_repo_path = os.path.join(self.parts_directory_path, "test")
    self.assertTrue(os.path.exists(main_repo_path))
    submodule_repo_path = os.path.join(main_repo_path, 'dir1',
                                       'submodule_repo')
    # Check if the submodule is not empty
    self.assertTrue(os.listdir(submodule_repo_path))

    # Get the head commit of the submodule repo
    head_commit_submodule_before_fetch = check_output(['git', 'rev-parse',
                                        'HEAD'],
                                        cwd=submodule_repo_path).strip()

    # Now, let's update the main repo as well as submodule repository.
    # Updating the repo will fetch the updated branch but doesn't checkout
    # to the updated master branch.
    recipe.update()
    head_commit_after_udpate_before_checkout = check_output(['git', 'submodule',
                                          'status', submodule_repo_path],
                                          cwd=main_repo_path).split()[0]

    # Checkout the submodule repository and get the head commit
    check_call(['git', 'checkout', 'master'], cwd=submodule_repo_path)
    head_commit_after_update_after_checkout = check_output(['git', 'rev-parse',
                                        'HEAD'],
                                        cwd=submodule_repo_path).strip()

    # Make another git clone of submodule and check the HEAD of the repo.
    submodule_recipe = self.makeGitCloneRecipe(
      {'repository': self.submodule_dir}
    )
    submodule_recipe.install()
    # Check the HEAD of the submodule
    head_commit = check_output(['git', 'rev-parse', 'HEAD'],
                                cwd=main_repo_path).strip()

    self.assertEqual(head_commit_after_update_after_checkout, head_commit)

def test_suite():
  suite = unittest.TestSuite((
      doctest.DocFileSuite(
          os.path.join(os.path.dirname(__file__), '..', '..', '..', 'README.rst'),
          module_relative=False,
          setUp=setUp,
          tearDown=zc.buildout.testing.buildoutTearDown,
          optionflags=optionflags,
          checker=renormalizing.RENormalizing([
              zc.buildout.testing.normalize_path,
              (re.compile(r'http://localhost:\d+'), 'http://test.server'),
              # Clean up the variable hashed filenames to avoid spurious
              # test failures
              (re.compile(r'[a-f0-9]{32}'), ''),
              ]),
          ),
      unittest.makeSuite(GitCloneNonInformativeTests),
      ))
  return suite

if __name__ == '__main__':
  unittest.main(defaultTest='test_suite')
